#! /usr/bin/perl -w

# 
# -------------------------------------------------------------------
#                                   X-BONE
#
#                          http://www.isi.edu/xbone
#                USC Information Sciences Institute (USC/ISI)
#                   Marina del Rey, California 90292, USA
#                          Copyright (c) 1998-2005
# 
# -------------------------------------------------------------------
#
# Copyright (c) 1998-2005 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation in source and binary forms for non-commercial
# purposes and without fee is hereby granted, provided that the above
# copyright notice appear in all copies and that both the copyright
# notice and this permission notice appear in supporting
# documentation, and that any documentation, advertising materials,
# and other materials related to such distribution and use acknowledge
# that the software was developed by the University of Southern
# California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived
# from this software without specific prior written permission.
# 
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT
# THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS
# PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so
# noted when applicable.
#
# -------------------------------------------------------------------
#
# Effort partly sponsored by the Defense Advanced Research Projects
# Agency (DARPA) and Air Force Research Laboratory, Air Force Materiel
# Command, USAF, under agreement numbers F30602-98-1-0200 (X-Bone) and
# F30602-01-2-0529 (DynaBone). The views and conclusions contained
# herein are those of the authors and should not be interpreted as
# necessarily representing the official policies or endorsements,
# either expressed or implied, of the Defense Advanced Research
# Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
#
# This work was partly supported by the NSF STI-XTEND (ANI-0230789)
# and NETFS (ANI-0129689) projects. Any opinions, findings, and
# conclusions or recommendations expressed in this material are those
# of the authors and do not necessarily reflect the views of the
# National Science Foundation.
#
# ----------------------------------------------------------------------------
#  $RCSfile: barp-table.pl,v $
#
# $Revision: 1.4 $
#   $Author: pingali $
#     $Date: 2005/03/31 07:03:44 $
#    $State: Exp $
# ----------------------------------------------------------------------------
# $Log: barp-table.pl,v $
# Revision 1.4  2005/03/31 07:03:44  pingali
# Updated copyright
#
# Revision 1.3  2004/01/22 23:48:14  rzhou
# add copyright
#
# Revision 1.2  2002/12/10 23:12:42  larse
# Merge Sunhee's PRM stuff from the BRANCH_FTN_PI_JUL2002 branch.
#
# Revision 1.1.2.2  2002/07/05 22:32:02  yushunwa
# Fixed some comment regarding usage.
#
# Revision 1.1.2.1  2002/07/05 22:00:42  yushunwa
# Renamed from outer.pl, script to generate barp tables.
#
# Revision 1.1  2002/01/12 04:17:01  yushunwa
# Script to combine innerlay barp tables into a single outerlay barp table.
#
# ----------------------------------------------------------------------------
#
# Temporary hack to gather barp table entries of several innerlays to construct
# a barp table for the outerlay.
#
# Usage: barp-table.pl ilay-ovl1 ilay-ovl2 ilay-ovl3
# where ilay-ovl* are the barp tables foe the corresponding innerlays.
#

use strict;
use Getopt::Long;
use Data::Dumper;
use FileHandle;

# clear path for taint mode
$ENV{PATH} = "";



# Constants
#
use constant IFCONFIG  => "/sbin/ifconfig "; # path to ifconfig command
use constant IPFW      => "/sbin/ipfw -q ";  # path to firewall comand



# Command line options
#
my @innerlays = ();      # array of innerlay addresses to multiplex over
my $outerlay  = "";      # IP address + netmask of the outerlay interface
my $verbose   = 0;       # flag: verbose diagnostics
my $debug     = 0;       # flag: debug inbound validation
my $rule      = 1;       # ipfw rule number for divert
my $port      = 55555;   # ipfw port number for divert
my $tundev    = "tun0";  # tun decive to use
my $help      = 0;       # print help and exit


my ($ovlname, $real_ips, $virtual_ips, @real_ips, @virtual_ips);
my %lists;
my @files = @ARGV;
my %barp;
my ($r, $v);
my (%outerlay_ips, $outerlay_barp);
my $run_dir = "/var/xbone/apps";
#my $run_dir = ".";
my $barpfile;

if($#ARGV==-1){
  print "No file specified.\n";
  exit;
}else{
  my $err=0;
  for my $f (@files){
    my $ovlname = $f;
    $ovlname =~ s/^\S*ilay-(\S+)/$1/;
    print "OVERLAY = $ovlname\n";
    if(not defined $outerlay_barp){
      $outerlay_barp = "barp-$ovlname";
    }
    open (IFILE, $f) or warn "Cant open $f: $!\n";
    if( $real_ips = <IFILE>){
      print "REAL_IPS = $real_ips";
    }else{
      $err=1;
      print "*** Mistake! Nothing in $f\n";
    }
    if( $virtual_ips = <IFILE>){
      print "VIRT_IPS = $virtual_ips";
    }else{
      $err=1;
      print "*** Mistake! Only 1 line in $f\n";
    }
    unless ($err){
      @real_ips = split /\s/, $real_ips;
      @virtual_ips = split /\s/, $virtual_ips;
      if (@real_ips != @virtual_ips){
        print "*** Error! Numbers of entries mismatched.\n";
      }else{
        while (($r = shift @real_ips) && ($v = shift @virtual_ips)){
	  if(not defined $outerlay_ips{$r}){
	    my $outer = $v;
	    $outer =~ s/^(\d+)\./10\./;
	    $outerlay_ips{$r}=$outer;
	  }
	  $barp{$outerlay_ips{$r}}{$ovlname} = $v;
	}
      }
      $err=0;
    }
  }
}

$Data::Dumper::Purity = 1;
my $testref = \%barp;
my $barp = Data::Dumper->Dump([$testref],["*barp"]);

print "barpfile: $outerlay_barp\n";
print "==========================================================\n";
print "$barp";
print "==========================================================\n";

my $fh = new FileHandle;  
if (!$fh->open ("$run_dir/$outerlay_barp", ">")) {
  die "! Error open $run_dir/$outerlay_barp: $!";
};

if (!chmod (0600, "$run_dir/$outerlay_barp")){
  warn "! Couldn't chmod $run_dir/$outerlay_barp to 0600: $!";
}

$fh->print ("$barp");
$fh->close;




