#! /usr/bin/perl -w

# 
# -------------------------------------------------------------------
#                                   X-BONE
#
#                          http://www.isi.edu/xbone
#                USC Information Sciences Institute (USC/ISI)
#                   Marina del Rey, California 90292, USA
#                          Copyright (c) 1998-2005
# 
# -------------------------------------------------------------------
#
# Copyright (c) 1998-2005 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation in source and binary forms for non-commercial
# purposes and without fee is hereby granted, provided that the above
# copyright notice appear in all copies and that both the copyright
# notice and this permission notice appear in supporting
# documentation, and that any documentation, advertising materials,
# and other materials related to such distribution and use acknowledge
# that the software was developed by the University of Southern
# California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived
# from this software without specific prior written permission.
# 
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT
# THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS
# PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so
# noted when applicable.
#
# -------------------------------------------------------------------
#
# Effort partly sponsored by the Defense Advanced Research Projects
# Agency (DARPA) and Air Force Research Laboratory, Air Force Materiel
# Command, USAF, under agreement numbers F30602-98-1-0200 (X-Bone) and
# F30602-01-2-0529 (DynaBone). The views and conclusions contained
# herein are those of the authors and should not be interpreted as
# necessarily representing the official policies or endorsements,
# either expressed or implied, of the Defense Advanced Research
# Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
#
# This work was partly supported by the NSF STI-XTEND (ANI-0230789)
# and NETFS (ANI-0129689) projects. Any opinions, findings, and
# conclusions or recommendations expressed in this material are those
# of the authors and do not necessarily reflect the views of the
# National Science Foundation.
#
#-----------------------------------------------------------------------------

use strict;

use HTTP::Daemon;
use HTTP::Status;
use HTTP::Response;

my %attack;

my $max_innerlays =  200;
my $max_pkts      =   50;
my $max_pkt_len   = 1500;


sub make_response () {
  # Make dummy status reply in PRM format, which consists of:
  #
  #   * a list of attacked innerlays
  #
  #       * zero or more integers (greater or equal to zero)
  #
  #   * zero or more packet descriptions, which are pairs consisting of:
  #
  #       * an innerlay number (integer greater or equal to zero)
  #       * the IP packet length in bytes
  #       * a valid SVG color name, see http://www.w3.org/TR/SVG/types.html
  #
  # The packet descriptions describe all packets the PRM routed since
  # the last status reply was generated. The list in the status reply
  # message is ordered by time, with the oldest packet description
  # first in the message, and the newest last. Colors should be used to
  # indicate packet type, with the exact scheme up to the PRM.
  #
  # The different pieces of the message are white-space separated. An
  # example message would be:
  #
  # 6 8 2 3 : 0 40 red 3 1460 black 2 1460 blue 0 1460 blue 1 40 red

  my @ALL_colors = qw(aliceblue antiquewhite aqua aquamarine azure beige
  bisque black blanchedalmond blue blueviolet brown burlywood
  cadetblue chartreuse chocolate coral cornflowerblue cornsilk crimson
  cyan darkblue darkcyan darkgoldenrod darkgray darkgreen darkgrey
  darkkhaki darkmagenta darkolivegreen darkorange darkorchid darkred
  darksalmon darkseagreen darkslateblue darkslategray darkslategrey
  darkturquoise darkviolet deeppink deepskyblue dimgray dimgrey
  dodgerblue firebrick floralwhite forestgreen fuchsia gainsboro
  ghostwhite gold goldenrod gray grey green greenyellow honeydew
  hotpink indianred indigo ivory khaki lavender lavenderblush
  lawngreen lemonchiffon lightblue lightcoral lightcyan
  lightgoldenrodyellow lightgray lightgreen lightgrey lightpink
  lightsalmon lightseagreen lightskyblue lightslategray lightslategrey
  lightsteelblue lightyellow lime limegreen linen magenta maroon
  mediumaquamarine mediumblue mediumorchid mediumpurple mediumseagreen
  mediumslateblue mediumspringgreen mediumturquoise mediumvioletred
  midnightblue mintcream mistyrose moccasin navajowhite navy oldlace
  olive olivedrab orange orangered orchid palegoldenrod palegreen
  paleturquoise palevioletred papayawhip peachpuff peru pink plum
  powderblue purple red rosybrown royalblue saddlebrown salmon
  sandybrown seagreen seashell sienna silver skyblue slateblue
  slategray slategrey snow springgreen steelblue tan teal thistle
  tomato turquoise violet wheat white whitesmoke yellow yellowgreen);

  my @colors = qw(blue cyan green magenta orange red tan yellow);

  my $msg = " ";
  foreach my $i (keys %attack) {
    $msg .= "$i ";
  }
  $msg .= ": ";
  my $tuples = int rand $max_pkts;
  for (my $i = 0; $i < $tuples; $i++) {
    # don't generate packets for attacked innerlays
    my $innerlay = int rand $max_innerlays;
    while (exists $attack{$innerlay}) {
      $innerlay = int rand $max_innerlays;
    }
    my $length = 1 + int rand ($max_pkt_len - 1);
    my $color = $colors[int rand $#colors];
    $msg .= "$innerlay $length $color ";
  }

  return $msg;
}


my $d = new HTTP::Daemon LocalPort => 50159, ReuseAddr => 1;
print "Please contact me at: ", $d->url, "\n";

while (1) {
while (my $c = $d->accept) {
  while (my $r = $c->get_request) {

    print $c->peerhost . ":" . $c->peerport . " " .
      $r->method . " " . $r->uri . "\n";

    if ($r->method eq "GET") {
      if ($r->url->path eq "/data") {
	my $msg = make_response;
	my $res = new HTTP::Response;
	$res->content($msg);
	$res->code(RC_OK);
	$c->send_response($res);
#	print "$msg\n";
      } else {
	if ($r->url->path eq "/") { $r->url->path("/index.html"); }
	my $file = substr $r->url->path, 1;
	$c->send_file_response($file);
      }

    } elsif ($r->method eq "POST") {
      my ($url, $param) = split "\\?", $r->uri;
      my ($tag, $val) = split "=", $param;

      if ($url eq "/attack") {
	$attack{$val} = 1;
	$c->send_response;

      } elsif ($url eq "/ok") {
	delete $attack{$val};
	$c->send_response;

      } else {
	$c->send_error(RC_FORBIDDEN)
      }

    } else {
      $c->send_error(RC_FORBIDDEN)
    }
  }
  $c->close;
  undef($c);
}
}
